title	"CLOCK program"

;CMOS = 1

public	pcfreq,vidram,count,fract,sec,min,hour,pic2old,msg

public	bcdbin,blank,clock,colon,dspclk,start,updtmr,vstore

EOI	= 20h		;End Of Interrupt command
pictrl	= 20h		;8259A interrupt controller control port
picmsk	= 21h		;Interrupt mask port
pictrl2	= 0a0h		;Interrupt controller port 2
picmsk2	= 0a1h		;Interrupt mask port 2

ESCAPE	=	27
CLKOFF	equ	70*2		;Video row 0, column 70

comment |This program defines a BIOS-driven clock routine that displays the
time of day at the upper right corner of the screen.  The time must be
calculated from the timer high and low words, since DOS is not reentrant.
|

bseg	segment	at 40h
	org	6ch
tickcnt	dd	?	;18.2-Hz tick count since midnite
bseg	ends

dseg	segment
pcfreq	dd	1193182		;8254-timer ticks/second
vidram	dw	0b000h		;Video RAM segment
count	dw	?		;Low-order word of sec-since-midnite count
fract	db	0,0		;Use 1,1 for 2-HZ CMOS interrupt
sec	db	0,59		;Seconds, max-seconds
min	db	0,59		;Minutes, max-minutes
hour	db	0,23		;Hours, max-hours
NTINT	equ	($ - sec)/2	;# time intervals (could add days, etc.)
pic2old	db	?		;Save loc for picmsk2
msg	db	"Type Esc key to quit$"
dseg	ends

sseg	segment	stack
	dw	100h dup(?)
sseg	ends

cseg	segment
	assume	cs:cseg

.386

comment |CLOCK - clock interrupt handler that displays time of day if interrupt is working OK
|

prvint	dd	?		;Previous interrupt

clock:	sti
	push	ds
	push	eax
	push	bx
	push	cx
	push	edx
ifdef	CMOS
INTVCT	equ	70h
	mov	al,EOI		;Reenable both PICs
	out	pictrl,al
	out	pictrl2,al
	mov	al,0ch		;Clear interrupt
	call	readp
	mov	ax,dseg
	mov	ds,ax
	assume	ds:dseg
	mov	cx,NTINT+1	;Increment up to NTINT+1 time intervals
	mov	bx,offset fract	;Point to smallest time interval counter
else
INTVCT	equ	1ch
	mov	ax,bseg		;This code executed about 18.2 times/second
	mov	ds,ax
	assume	ds:bseg		;Set edxeax = timer*65536, i.e.,
	mov	eax,tickcnt-2	; clock tick count before division by 65536
	movzx	edx,word ptr tickcnt+2

	push	dseg
	pop	ds
	assume	ds:dseg

	div	pcfreq		;eax = # seconds since midnight
	cmp	ax,count	;Change?				
	jz	short clock8	;No
	mov	count,ax	;Update sec, min, hour and dispatch
	mov	cx,NTINT	;Increment up to NTINT time intervals
	mov	bx,offset sec	;Point to smallest time interval counter
endif
	call	rtc		; real-time-clock functions
clock8:	pop	edx
	pop	cx
	pop	bx
	pop	eax
	pop	ds
	jmp	[prvint]	;Chain to previous interrupt


comment |RTC - maintain time of day values and dispatch operations at periodic intervals.  ax changed
|

rtc:	mov	al,[bx]		;Get next time interval
	inc	byte ptr [bx]	;Increment it
	cmp	al,[bx+1]	;Complete cycle for this interval?
	jb	short rtc2	;If not, leave loop
	mov	byte ptr [bx],0	;Yes.  Zero value and go on to next
	inc	bx		;Advance to next time interval counter
	inc	bx
	loop	rtc		;Do all time intervals, if necessary


;Real-time-clock routines.  bx points to the time interval being incremented.
;Replace nops with desired code to be executed periodically.

rtc2:	nop			;This code executed every fract
	cmp	cl,4
	jz	short dspcl9	;(rz) Don't display clock
	nop			;This code executed every second
	cmp	cl,3
	jz	short rtc6
	nop			;This code executed every minute
	cmp	cl,2
	jz	short rtc6
	nop			;This code executed every hour
	cmp	cl,1
	jz	short rtc6
				;This code executed every day
rtc6:				;Fall thru to dspclk (if you want)

dspclk:	push	es		;Display clock at upper-right on screen
	push	di
	mov	es,vidram
	mov	di,CLKOFF
	call	blank		;Display some leading space
	call	blank
	mov	al,hour
	call	dspdec		;Display hours
	call	colon
	mov	al,min
	call	dspdec		;Display minutes
	call	colon
	mov	al,sec
	call	dspdec		;Display seconds
	pop	di
	pop	es
dspcl9:	ret

dspdec:	aam			;(ah,al) = (al/10, remainder)
	or	ax,"00"		;Convert to ASCII
	push	ax		;Display al in BCD format
	mov	al,ah
	call	vstore		;Display it
	pop	ax

vstore:	mov	ah,7		;Attribute byte
	stosw
	ret

blank:	mov	al," "
	jmp	vstore

colon:	mov	al,":"
	jmp	vstore


comment |START - main program to test clock interrupt
|

start:	mov	ax,dseg
	mov	ds,ax		;ds:dseg
	mov	dx,offset msg	;Display how to quit
	mov	ah,9
	int	21h
	mov	ah,15
	int	10h
	cmp	al,7		;Monochrome?
	jz	short start2
	mov	vidram,0b800h	;No

start2:	call	updtmr		;Update BIOS timer_low/high
	mov	ah,2ch		;ch/cl/dh/dl = hours/mins/secs/hundredths
	int	21h
	mov	sec,dh
	mov	min,cl
	mov	hour,ch
	mov	ax,3500h+INTVCT	;Save previous interrupt
	int	21h
	mov	word ptr prvint,bx
	mov	word ptr prvint+2,es	
	push	ds
	push	cs
	pop	ds
	mov	dx,offset clock	;ds:clock (ds = cs in .com file)
	mov	ah,25h		;Take over interrupt
	int	21h
	pop	ds		;ds:dseg
ifdef	CMOS
	call	eblcms		;Enable CMOS timer interrupt
	mov	ax,2f0ah
	call	writep
endif

start4:	mov	ax,0		;Wait for keyboard interrupt
	int	16h
	cmp	al,ESCAPE	;Wait for ESC
	jnz	start4
ifdef	CMOS
	call	dblcms		;Disable CMOS timer interrupt
endif
	lds	dx,prvint	;Restore previous interrupt vector
	mov	ax,2500h+INTVCT	;Restore previous interrupt
	int	21h
	mov	ax,4c00h
	int	21h		;Quit
	

ifdef	CMOS

comment |READP - return al = current value of cmsadr given by entry value of
al. al changed.
|

readp:	out	70h,al		;Choose the requested cmsadr
	jmp	$+2		;Delay in case cmos not so fast
	in	al,71h		;al = current value
	ret


comment |DBLCMS/EBLCMS - disable/enable CMOS timer interrupt
|

dblcms:	mov	al,pic2old	;Restore old interrupt mask
	out	picmsk2,al
	mov	al,0bh		;Read CMOS B register
	call	readp
	and	al,not 40h	;Disable periodic interrupt
	jmp	short eblcm2

eblcms:	in	al,picmsk2	;Get 2nd 8259 interrupt mask
	mov	pic2old,al	;Save it for restoration on quitting
	and	al,not 1	;Enable RTC interrupts
	jmp	$+2
	out	picmsk2,al
	mov	al,0bh		;Read CMOS B register
	call	readp
	or	al,40h		;Enable periodic interrupt
eblcm2:	mov	ah,al
	mov	al,0bh		;Fall thru to writep


comment |WRITEP - output value ah to cmsadr al.  al changed.
|

writep:	out	70h,al		;Output to cmos cmsadr al
	jmp	$+2
	mov	al,ah		;Output value ah
	out	71h,al
	ret
endif

comment |BCDBIN - return ax = binary number corresponding to packed BCD number
in al. ax changed.
|

bcdbin:	mov	ah,al		;Unpack the nibbles
	shr	ah,4
	and	al,0fh
	aad			;ax = ah*10 + al
	ret


comment |UPDTMR - update the BIOS RAM timer count at 40:6ch using the CMOS RTC
value.  Return ch/cl/dh = hours/mins/secs.  eax, cx, edx changed.
|

updtmr:	mov	ah,2		;Get al/ch/cl/dh = BCD hours/hours/mins/secs
	int	1ah		;Takes care of effectively latching the time
	mov	al,dh
	call	bcdbin		;Convert BCD sec to binary sec
	cwde			;Zero high word of eax
	mov	edx,eax		;Save the seconds
	mov	al,cl		;al = minutes
	call	bcdbin
	imul	ax,ax,60	;ax = number of seconds in the minutes
	add	dx,ax		;dx = seconds + 60*minutes
	mov	al,ch		;al = bcd hour
	call	bcdbin
	imul	eax,eax,3600	;eax = 60*60*hours
	add	eax,edx		;eax = # seconds in day up to present time
	mul	pcfreq		;edxeax = 1193182-Hz tick count
	shrd	eax,edx,16	;Divide by 65536 to get 18.2-Hz tick count
	push	ds		;Store it in BIOS RAM tick count
	push	bseg		;(Can also use ah=1/int-1ah with desired
	pop	ds		; 18.2-Hz tick count given by cxdx)
	assume	ds:bseg
	mov	tickcnt,eax
	pop	ds
	ret

cseg	ends

	end	start
