title	cmos.asm - program to examine/modify CMOS RAM

;(c) copyright 1994 by Scroll Systems, Inc.

comment |CMOS - program to store a byte value into PC CMOS RAM and adjust
check sum accordingly.  Usage:

	cmos port [value]

where port and value are given in hex.  If the value is missing, the current
value is displayed.  For example,

	cmos 10 24

stores the value 24h into the cmos memory location 10h, which is the diskette
drive type configuration byte.  The value 24h specifies type 2 for drive A
(1.2 MB 5.25") and type 4 for drive B (1.44 3.5").  Having type this, the
command

	cmos 10

would display

	Current value = 24

Running cmos without arguments displays the current CMOS RAM contents.
The commands

cmos ?
cmos help

both display a message on how to use cmos.com

IMPORTANT: This program must be run as a com file.
|

public	bcdbin, blk
public	cmos, cmsadr, co, coalsp, conv, cow, crlf
public	dspcms
public	lbyte
public	nibble
public	outbx
public	readp
public	strchk
public	tom
public	writep

include	return.inc

CR	equ	0dh
LF	equ	0ah

cseg	segment
	assume	cs:cseg, ds:cseg

	org	5dh
fcb1	db	?		;First File Control Block
	org	6dh
fcb2	db	?		;Second File Control Block

	org	100h		;com file

cmos:	mov	dx,offset msg
	mov	ah,9
	int	21h
	mov	ax,word ptr fcb1
	call	bcdbin		;dx = CMOS RAM address
	jnc	short cmos4	;Go use it
	mov	si,offset usemsg
	cmp	al,'?'		;User want help?
	jz	short cmos3	;Yep
	cmp	ax,"EH"		;HELP?
	jnz	short dspcms	;No, go display current CMOS contents

cmos3:	call	tom		;Display message *si and return to DOS
	return			;(Initialize return macros)

cmos4:	mov	cmsadr,dl	;Save CMOS RAM address
	mov	ax,word ptr fcb2
	call	bcdbin		;dx = value
	mov	al,cmsadr
	mov	bl,al
	call	readp		;al = current value of cmsadr al
	jnc	short cmos6	;Go output new value

	call	lbyte		;No value specified. Convert current value to
	mov	valmsg2,ax	; ASCII and store it in message
	mov	si,offset valmsg
	jmp	cmos3		;Go display message with current value

cmos6:	xchg	bx,ax		;bl = current value for adjusting check sum
	mov	ah,dl		;ah = value to output to requested cmsadr (al)
	call	writep		;Write ah to cmos port al
	cmp	cmsadr,10h	;cmsadr within the range 10h to 2dh?
	rc			;No
	cmp	cmsadr,2eh
	rnc			;No
				;Yes: need to adjust checksum
	sub	al,bl
	cbw
	xchg	bx,ax		;bx = new value - old value
	mov	al,2eh
	call	readp
	mov	ah,al		;ah = high byte of current checksum
	mov	al,2fh
	call	readp		;ax = current checksum
	add	bx,ax		;bx = new checksum.  Fall thru to strchk


comment |STRCHK - store cmos checksum
|

strchk:	mov	al,2eh		;Output high byte to checksum
	mov	ah,bh
	call	writep
	mov	al,2fh		;Output low byte to checksum
	mov	ah,bl
	jmp	writep


comment |DSPCMS - display CMOS RAM contents
|

dspcms:	xor	bx,bx		;Zero the checksum
	xor	di,di		;di = location counter
	mov	si,offset ramsg	;*si = list of messages and codes
	
dspcm2:	mov	ax,di		;Display the current CMOS location
	call	coalsp		; followed by a space
	call	tom		;Display next message
	mov	ax,' :'
	call	cow		;Display ": "
	lodsb			;|al| = # bytes to display
	cbw			;Set sign plus.  Can you figure this out?
	xor	al,ah		; How is it better than
	sub	al,ah		; or al,al   jns $+4    neg al ?
	cbw			;ah = 0
	xchg	cx,ax		;cx = # bytes to display

dspcm4:	mov	ax,di		;*ax = next value
	call	readp		;al = *cmos:di
	test	byte ptr [si-1],0ffh	;code < 0?
	jns	short dspcm6	;No
	add	bx,ax		;Add to checksum
dspcm6:	call	coalsp
	inc	di		;*cmos:di = next value
	loop	dspcm4
	call	CRLF		;Display CarriageReturn/LineFeed
	test	byte ptr [si],0ffh
	jnz	dspcm2		;More to go
	inc	si		;Bypass terminating null at end of list
	call	tom		;"Checksum: "
				;Fall thru to display bx in hex


;Utility routines follow

comment |OUTBX - Display bx in hex.  ax changed.
|

outbx:	mov	al,bh		;Display high byte first
	call	lbyte		;Convert al to ASCII ax
	call	cow		;Display ax (al first)
	mov	al,bl		;Now low byte
	call	lbyte		;Fall thru to cow


comment |COW - display the word al ah.  al changed.
|

cow:	call	co
	mov	al,ah		;Fall thru to co


comment |CO - console output al to screen.  No regs changed (except sometimes
bp...)
|

co:	push	ax
	push	bx
	mov	ah,0eh
	cmp	al,9		;Tab?
	jnz	short co2
	mov	al," "
co2:	xor	bx,bx		;Display RAM page 0
	int	10h		;Slow, but it works
	pop	bx
	pop	ax
	return


comment |COALSP - co the binary al followed by a space (blank)
|

coalsp:	call	lbyte		;Convert to ASCII
	call	cow		;Display ax


comment |BLK - display a blank
|

blk:	mov	al,' '
	jmp	co


comment |CRLF - Display Carriage-Return Line-Feed. al changed.
|

crlf:	mov	al,CR
	call	co
	mov	al,LF		;Fall thru to co
	jmp	co


comment |TOM - "Terminal Output Message" - from old Zapple days.  Display
string [si] up to null.  ax, si changed.
|

tom:	lodsb			;Display message [ds:si]. Get a character
	or	al,al
	rz
	call	co		;Output char
	jmp	tom


comment |BCDBIN RNC with dx = binary version of ASCII bytes al/ah.  al changed
unless no conversion (RC).
|

bcdbin:	xor	dx,dx		;Zero the result
	call	nibble		;al = high or only hexadigit
	rc			;No hex value
	mov	dl,al		;dl = high or only hexadigit binary value
	mov	al,ah
	call	nibble		;al = low hexadigit
	cmc
	rnc			;RNC.  dx = one-digit value
	mov	cl,4		;Move over for low hexadigit binary value
	shl	dx,cl		;dx = 16*(high hexadigit)
	add	dl,al		;dx = two-hexadigit binary value
	return


comment |NIBBLE - Convert ASCII al to hex low nibble. Set CY if non-hex input.
Reverse of CONV. al changed.
|

nibble:	cmp	al,'0'
	rc			;Set CY if <'0'
	cmp	al,'9'
	ja	short nibbl2	;Not a digit
	sub	al,'0'
	return			;RNC with hex binary value
nibbl2:	cmp	al,'A'		;Not UC alphabetic
	rc
	cmp	al,'G'
	cmc
	rc			;Not hex
	sub	al,'A'-10d
	return			;RNC with hex binary value


comment |LBYTE - return ax = ASCII representation of hex number in al.  ax
changed.
|

lbyte:	mov	ah,al
	call	conv
	xchg	ah,al		;Low nibble gets displayed second
	shr	al,1
	shr	al,1
	shr	al,1
	shr	al,1


comment |CONV - Convert low hex nibble of al to ASCII. al changed.
|

conv:	and	al,0fh		;Ensure a single hex digit
	add	al,90h		;add 90H using BCD arithmetic
	daa
	adc	al,40h		;then 40H plus CY
	daa
	return			;Puzzle: figure out how this routine works


comment |READP - return al = current value of cmsadr given by entry value of
al. al changed.
|

readp:	out	70h,al		;Choose the requested cmsadr
	jmp	$+2		;Delay in case cmos not so fast
	in	al,71h		;al = current value
	return


comment |WRITEP - output value ah to cmsadr al.  al changed.
|

writep:	out	70h,al		;Output to cmos cmsadr al
	jmp	$+2
	mov	al,ah		;Output value ah
	out	71h,al
	return


cmsadr	db	?
msg	db	"cmos 94. (c) copyright 1994 by Scroll Systems, Inc."
	db	CR,LF,'$'
usemsg	db	"USAGE:  cmos [cmosadr [value]] | ? | help",CR,LF,LF
	db	"where cmosadr and optional value are in hex. If value is "
	db	"missing, current",CR,LF
	db	"value is displayed. Else value is stored in cmosadr "
	db	"and checksum adjusted",CR,LF
	db	"appropriately.   Running cmos without arguments displays "
	db	"current CMOS RAM",CR,LF
	db	"contents.  This message is displayed by either of the "
	db	"following commands:",CR,LF
	db	"    cmos ?",CR,LF
	db	"    cmos help",CR,LF,0

valmsg	db	"Current value = "
valmsg2	dw	?
	db	0


;CMOSAddressName,0,code
;|code| = # bytes of field; code < 0 ==> include in checksum

ramsg	db	"Seconds/Seconds alarm",0,2		;00h-01h
	db	"Minutes/Minutes alarm",0,2		;02h-03h
	db	"Hour   /Hour alarm   ",0,2		;04h-05h
	db	"Day of week          ",0,1		;06h
	db	"Day/Month/Year       ",0,3		;07h-09h
	db	"Status Regs A,B,C,D  ",0,4		;0ah-0dh
	db	"Diagnostic/ResetCode ",0,2		;0eh-0fh
	db	"DisketteType/Reserved",0,-2		;10h-11h
	db	"FixedDiskType/Resrved",0,-2		;12h-13h
	db	"Equipment installed  ",0,-1		;14h
	db	"Kbytes base RAM      ",0,-2		;15h/16h (low/high)
	db	"Kbytes extended RAM  ",0,-2		;17h/18h (low/high)
	db	"Drive C/D extend type",0,-2		;19h-1ah
	db	"Reserved ",0,-18			;1bh-2ch
	db	"Additional flags     ",0,-1		;2dh
	db	"Checksum             ",0,2		;2eh/2fh (high/low)
	db	"KBytes RAM over 1 MB ",0,2		;30h/31h (low/high)
	db	"Century              ",0,1		;32h
	db	"System info          ",0,1		;33h
	db	"Reserved             ",0,12		;34h-3fh
	db	0					;Done with list
	db	"Checksum: ",0

cseg	ends

	end	cmos
