cpuid	macro			;Generate Pentium cpuid instruction
	db	0fh
	db	0a2h
	endm

dseg	segment
cputyp	db	?
winvno	dw	0		;Windows version #
dpmivno	dw	0		;DPMI major/minor version number
XMSctrl	dd	?		;XMS control addr or DPMI mode-switch addr
XMSvern	dw	0		;XMSvern
XMSfree	dw	0		;#K in largest XMS XRAM block
mdswtch	dd	?		;Mode-switch address
gdtadr	df	?
err386	db	"Need 386 or bigger CPU$"
errDPMI	db	"Need to run under Windows DPMI$"
hello	db	"Hi there from Windows DPMI protected mode",0dh,0ah
	db	"Master/slave 8259 address = $"

dseg	ends

sseg	segment	stack
	dw	100 dup(?)
sseg	ends

psp	segment	at 0
	org	2
topmem	dw	?
psp	ends

cseg	segment
	assume	cs:cseg, ds:dseg, ss:sseg

.386p
start:	push	es		;Save psp segment
	mov	ax,dseg
	mov	ds,ax		;ds:dseg
	call	getcpu
	mov	cputyp,al
	cmp	al,3		;At least 386?
	mov	dx,offset dseg:err386
	jc	short start3	;No, go display error msg and quit
	smsw	ax		;CPU running in protected mode?
	test	al,1
	jnz	short start4	;Yes
				;No, could use real big mode...
start2:	mov	dx,offset dseg:errDPMI

start3:	mov	ah,9		;Display error message
	int	21h
	mov	ax,4c01h
	int	21h		;Bye


;CPU in protect mode, must be V86

start4:	call	chkXMS		;Get XMS version # (uses XMSctrl)
	jnz	start2		;Not there; display error msg and quit
	mov	ax,1600h
	int	2fh		;Get Enhanced-Mode Windows Installed State
	test	al,7fh
	jz	start2		;386 enhanced-mode Windows not running
	mov     winvno,ax
	mov	ax,1687h	;DPMI there?
	int	2fh
	or	ax,ax		;Succeed?
	jnz	start2		;No
	mov	dpmivno,dx
	mov	word ptr mdswtch,di	;Save call to switch into protect mode
	mov	word ptr mdswtch+2,es	; (returned by int 2fh)

	pop	es
	assume	es:psp
	mov	ax,topmem	;SST's topmem value
	sub	ax,si		;Subtract # para's needed by DPMI
	mov	es,ax		;Point to top DOS memory block
	assume	es:nothing
	xor	ax,ax		;Flags: 16-bit app
	call	[mdswtch]	;Switch to protect mode (first time)
	jc	start2		;Failed, go display error message and quit

	mov	dx,offset dseg:hello
	mov	ah,9		;We're in protected mode
	int	21h		;Brag about it a bit

	mov	ax,400h		;Get ax = version and other goodies
	int	31h		; (cl = processor type)
	push	dx		;Display 8259 master/slave base interrupt
	mov	al,dh		; numbers.  Master first
	call	pbyte
	mov	dl,'/'
	call	print2
	pop	ax
	call	pbyte		;Then slave
	mov	ax,4c00h
	int	21h		;Then quit with normal-exit code


comment |CHKXMS - RNZ if XMS not active.  Else get XMS version # and address
|

chkXMS:	mov	ax,4300h	;Check to see if XMS driver is installed
	int	2fh
	cmp	al,80h		;Driver there?
	jnz	short chkXM9	;No: RNZ
	push	es		;Save es:di in case DPMI real-to-protect addr
	mov	ax,4310h	;Get driver address
	int	2fh
	mov	word ptr XMSctrl,bx	;Store it for XMS calls
	mov	word ptr XMSctrl+2,es
	pop	es
	xor	ax,ax
	call	[XMSctrl]
	mov	XMSvern,ax	;Save version number
	mov	ah,8		;Find out largest XRAM block available
	call	[XMSctrl]
	mov	XMSfree,ax	;# K-bytes of block
	cmp	ax,ax
chkXM9:	ret


comment |GETCPU - return al = 0 for 8088, = 1 for 80186, = 2 for 80286, = 3 for 80386.  ax, bx, cx changed
|

getcpu:	mov	bx,sp		;Determine machine type. Save sp
	mov	ax,sp		;ax = nonzero
	and	sp,not 3	;Be sure stack is on dword boundary for 486
	mov	cl,20h
	shr	ax,cl		;80186, 80286 shift modulo 20h, i.e., here nop
	jz	short getcp5	;8086 shifts 32 positions clearing ax register
	push	sp		;80286 and higher push sp value before push
	pop	ax
	test	al,2		;8086, 80186 would have decremented value
	mov	al,1		;At least 80186
	jnz	short getcp5
	inc	ax		;At least 80286
	sgdt	fword ptr gdtadr
	cmp	byte ptr gdtadr+5,0ffh	;80286?
	jz	short getcp5
	pushfd			;Save flags twice: First to restore them
	pushfd			;Second to get them
	pop	eax
	or	eax,240000h	;Set ID & Alignment Check (AC) bits in eflags
	push	eax
	popfd			;Tell CPU about it
	pushfd			;See if CPU took the bait
	pop	eax
	shr	eax,10h
	and	al,24h		;Isolate IC and AC flags
	jz	short getcp3	;Can't set either: must be 386
	cmp	al,4		;Can't set ID bit?
	jz	short getcp4	;No, must be 486
	xor	eax,eax		;586 or higher
	inc	ax		;eax = 1
	cpuid			;Get CPU ID
	mov	al,ah		;Forget about model and stepping numbers...
	and	al,0fh		;Isolate ID. 5 for Pentium
	jmp	short getcp4
getcp3:	mov	al,3		;386
getcp4:	popfd			;Restore flags

getcp5:	mov	sp,bx		;Restore sp in case and sp,not 3 changed it
	cbw			;ax = cputyp
	ret


comment	|PBYTE - display al in hex
|

pbyte:	push	ax	;Save a copy of al
	shr	al,4
	call	print	;Print hex digit
	pop	ax	;Restore original al
			;Fall thru to print 2nd
print:	and	al,0fh	;Mask off high nibble
	add	al,90h	;Add 90H using BCD arithmetic
	daa
	adc	al,40h	;Then 40H plus CY
	daa
	xchg	dx,ax	;Display with DOS character

print2:	mov	ah,2	; display entry
	int	21h
	ret

cseg	ends

	end	start
