title kb1.asm - program to adjust typematic rate/delay

;(c) copyright 1994 by Scroll Systems, Inc.

comment |This program sets the typematic rate/delay according to the keyboard
discussion in the IBM AT Technical Reference (see section "Set Typematic
Rate/Delay (Hex F3).  The program assembles into a COM file with the following
DOS-command line meaning:

kb1 [period [delay]]

where period and delay are optional parameters.  If none are specified, null values are used.  The period is given by a decimal number.  If none is given
(and hence no delay is specified either), a minimum period of 8 is
automatically assumed, which converts into a repeat rate of about 30/second. 
The maximum value is 31, which gives a repeat rate of about 2/second.

The delay is a single decimal digit between 0 and 3.  The actual delay is 1
plus this value times 250 milliseconds.
|

bseg	segment	at 40H
	org	97H
kbdflg4	db	?	;AT ROM BIOS keyboard flag
bseg	ends

cseg	segment
	assume	cs:cseg,ds:cseg

	org	5DH	;File 1 name
period	dw	?	;Used here for autorepeat period
	org	6DH	;File 2 name,
delay	db	?	;Used here for autorepeat delay

	org	100H
start:	mov	dx,offset signon	;Display signon message
	mov	ah,9
	int	21h
	cmp	byte ptr period,"?"	;Help message requested?
	jnz	short start1
	mov	dx,offset hlpinfo	;Yes. Display and quit
	int	21h
	ret

start1:	mov	dx,offset signend	;Finish signon
	int	21h
	mov	ax,period		;Construct delay/period byte
	and	al,0FH
	cmp	ah," "			;Second digit blank?
	jz	short start2		;Yes
	and	ah,0FH			;No, incorporate second digit
	mov	ch,ah
	mov	cl,0AH
	mul	cl
	add	al,ch

start2:	mov	ah,al			;ah = desired period
	mov	al,delay		;Incorporate delay in bits 5, 6
	and	al,3
	mov	cl,5
	shl	al,cl			;Shift it to bits 5,6
	or	ah,al			;ah = finished period/delay code

autort:	mov	bx,bseg			;Setup ds to address ROM BIOS RAM
	mov	ds,bx			; for keyout routine
	assume	ds:bseg

	mov	bh,0F3H			;Send typematic period/delay command
	call	keyout
	mov	bh,ah			;Send desired code
	call	keyout
	mov	bh,0F4H			;Enable keyboard


;KEYOUT - output bh to 8042 handling acknowledge and retries

keyout:	mov	bl,3			;Maximum of three retries
keyou1:	cli
	and	kbdflg4,0CFH		;Kill acknowledge and resend bits
	sub	cx,cx
keyou2:	in	al,64H			;Wait for last command to be accepted
	test	al,2			;Input buffer empty?
	loopnz	keyou2			;Fall through if keyboard gone
	mov	al,bh			;Send value. Should cause INT 9 to set
	out	60H,al			; acknowledge or resend flag bit
	sti
	mov	cx,1A00H		;Load count for 10 msec
keyou4:	test	kbdflg4,10H		;Acknowledge received?
	jnz	short keyou9
	test	kbdflg4,20H		;Loop until resend received
	loopz	keyou4
	dec	bl			;Decrement retry count
	jnz	keyou1
	or	kbdflg4,80H		;Turn on transmit error flag
keyou9:	ret


signon:
db "KB1 - (c) Scroll Systems, Inc., Tucson, AZ.  Freeware program to set",13,10
db "AT keyboard typematic rate/delay. $"

signend:
db "Type kb1 ? for help",13,10,"$"

hlpinfo:
db 13,10,10,"At the DOS command prompt, type",13,10,10
db "kb1 [period [delay]]",13,10,10
db "where period and delay are optional parameters.  If none are specified,",13,10
db "null values are used.  The period is given by a decimal number.  If none",13,10
db "is given (and hence no delay is specified either), a minimum period of 8",13,10
db "is automatically assumed, which converts into a repeat rate of about 30/second.",13,10
db "The maximum value is 31, which gives a repeat rate of about 2/second.",13,10,10
db "The delay is a single decimal digit between 0 and 3.  The actual delay is 1",13,10
db "plus this value times 250 milliseconds.",13,10,"$"

cseg	ends

	end	start
