;Include file for keyboard.asm.  Defines direct-screen console output (co)
;routines.  Can interpret a number of control characters or interpret only
;Esc sequences.  The latter approach is used with SST, which displays the Esc
;char as a reverse-video $.

public	co
public	scrldn, scrlup, setcsr
public	tom
public	vdinit

ESCAPE	equ	27
BOLD	equ	8
RVV	equ	7fh
UNDER	equ	2

bdata	segment	at 40h		;See PC Technical Reference BIOS listing
	org	10h
EquipFlag	label	byte
	org	17h
KbdFlag	label	byte		;RAM keyboard flag byte
	org	1ah
KbdHead	dw	?
KbdTail	dw	?
	org	49h
CRTmode	db	?		;Current CRT mode
CRTcols	db	?		;# cols on text screen
	org	50h
CursorPos	dw	?
 	org	60h
CursorMode	dw	?
	org	63h
Addr6845	dw	?
CRTModeSet	db	?
	org	6ch
TimerLow	dw	?
	org	84h
CRTrows	db	?		;Valid for EGA/VGA
	org	88h
cnfswtc	db	?		;Configuration switches
bdata	ends


;Control sequence dispatch table

	even
ctltbl	dw	specl,1		;^@ - special PC key code (code)
	dw	setatr,1	;^A - set attribute (attribute)
	dw	togbld,0	;^B - toggle bold
	dw	setcsr,2	;^C - set cursor position (col row)
	dw	dupchr,2	;^D - duplicate character (n char)
	dw	eblcrs,1	;^E - enable/disable hardware cursor (n)
	dw	0,0		;^F
	dw	bell,0		;^G - bell
	dw	bckspc,0	;^H - backspace
	dw	tab,0		;^I - tab
	dw	linefd,0	;^J - line feed
	dw	0,0		;^K
	dw	clrscn,0	;^L - erase screen
	dw	caretn,0	;^M - carriage return
	dw	togund,0	;^N - toggle underscore
	dw	0,0		;^O
	dw	0,0		;^P
	dw	0,0		;^Q
	dw	togrvv,0	;^R - toggle reverse video
	dw	sclrct,4	;^S - scroll rectangle (col row wdth hght)
	dw	0,0		;^T -
	dw	0,0		;^U -
	dw	0,0		;^V -
	dw	wiprct,4	;^W - wipe (erase) rect (col row wdth hght)
	dw	setcol,1	;^X - set col (n)
	dw	setrow,1	;^Y - set row (n)
	dw	0,0		;^Z -
	dw	co3,0		;^[ - Escape

NCTL	equ	($-ctltbl)/4

cursor	dw	?		;Screen RAM offset
vidram	dw	0b000h		;Default monochrome screen
dcursor	equ	dword ptr cursor
crtprt	dw	3b4h		;Monochrome 6845 cursor port
linel	dw	?		;Screen width
lpp	dw	25		;"lines per page" (# rows on screen)
csrcol	db	?		;Cursor column
csrrow	db	?		;Cursor row
ctlcnt	dw	0		;# control characters still needed
ctlstr	db	5 dup (?)	;Control sequence string
ctlcol	equ	byte ptr ctlstr+4
ctlrow	equ	byte ptr ctlstr+3
ctlwdth	equ	byte ptr ctlstr+2
ctlhght	equ	byte ptr ctlstr+1
attrib	db	7		;Screen attribute
nrmatr	db	7
dspflgs	db	?		;Display flags with meanings:

;bit 0 = 1	suppress display of hardware cursor
;bit 1 = 1	only recognize Esc sequences: display other control chars


comment |CO - console output routine
|

co:	mov	bx,ctlcnt	;Need more bytes for a control sequence?
	test	bx,bx
	jz	short co2	;No
	mov	ctlstr[bx],al	;Store control-sequence byte
	dec	ctlcnt		;One less
	jnz	short co6	;Return for more bytes
	mov	bl,ctlstr	;Command ready to dispatch
	mov	bh,0		;Point into control dispatch table
	shl	bx,1		;Entries consist of offset, number (2 dw's)
	shl	bx,1
co1:	call	ctltbl[bx]	;Dispatch control command
	jz	short co9	;No character to process

co2:	cmp	ax,ESCAPE	;Is character an Esc ?
	jne	short co36	;No
	mov	ctlstr,al	;Start Esc sequence
	mov	al,1		;Need 1 byte to identify which Esc sequence
	jmp	short co5

co3:	pop	ax		;Esc sequence comes here. Kill return to co1
	mov	al,ctlstr+1	;al = escape code
	test	al,al
	jns	short co4	;Codes 80h - 0ffh are duplication codes
	and 	al,7fh		;al = duplication count
	mov	ctlstr+2,al	;Store count in control sequence
	mov	ctlstr,'D'-40h	;Convert Esc command into duplication command
	mov	ax,1		;Set up to get character to be duplicated
	jmp	short co5

co36:	test	dspflgs,2	;Display control chars except for Esc?
	jne	short co8	;Yes (feature used by SST)

co4:	cbw
	cmp	ax,NCTL		;Control character?
	jnc	short co8	;No or above those handled here
	mov	ctlstr,al	;Save control char
	push	ax
	xchg	bx,ax		;Point into control dispatch table
	shl	bx,1		;Entries consist of offset, number (2 dw's)
	shl	bx,1
	mov	ax,ctltbl[bx]
	test	ax,ax		;Routine defined?
	pop	ax
	jz	short co8	;No, go store char in screen RAM
	mov	ax,ctltbl[bx+2]	;ax = control-sequence xtra char count
	test	ax,ax		;Need more chars in control sequence?
	jz	co1		;No, go dispatch control squence
co5:	mov	ctlcnt,ax	;Store # chars to get
co6:	return

co8:	call	vstore		;Store char in screen RAM
co9:	test	dspflgs,1	;Hardware cursor display suppressed?
	rnz			;Yes
				;Fall thru to discrs

comment |DISCRS - display hardware cursor
|

discrs:	mov	ax,cursor	;Display hardware cursor
	shr	ax,1
	push	ax		;Save low byte of character offset
	mov	ah,al
	mov	al,15d		;Output high offset byte to 6845 register 15d
	call	ot6845
	pop	ax
	mov	al,14d		;Output low offset byte to 6845 register 14d


comment |OT6845 - Direct Output ax to IBM 6845.  No regs changed.
|

ot6845:	push	dx
	mov	dx,crtprt
	call	outdx
	pop	dx
	return


comment |OUTDX - output ah to register al in device dx.  No regs changed.
|

outdx:	out	dx,al		;Select register
	jmp	$+2		;Slow down
	inc	dx
	xchg	al,ah
	out	dx,al		;Output value
	xchg	al,ah		;Restore ax, dx
	dec	dx
	return


;**************************Control character handlers*************************

comment |SPECL - ^@: return special PC key code: al = 0 and ah = std PC value,
e.g., ax = 3b00h for function key 1 (F1)
|

specl:	mov	ax,word ptr ctlstr
	or	ax,ax		;RNZ to indicate char to process
	return


comment |SETATR - ^A: set attribute
|

setatr:	mov	al,ctlstr+1
	mov	attrib,al
	cmp	al,al
	return


comment |TOGBLD - ^B: toggle bold attribute
|

togbld:	xor	attrib,BOLD
	cmp	al,al		;RZ
	return


comment |SETCSR - ^C: set cursor position to col row parameters
|

setcsr:	mov	dl,ctlstr+2	;dl = column
	mov	dh,ctlstr+1	;dh = row.  Fall thru to setcs1


comment |SETCS1 - Set cursor = dh*linel + 2*dl
|

setcs1:	mov	word ptr csrcol,dx
	push	ax
	push	dx
	mov	ax,linel
	neg	ax
	cmp	dh,0ffh		;Line before screen?
	jz	short setcs2
	neg	ax
	push	dx
	mov	dl,dh
	mov	dh,0
	mul	dx
	pop	dx
	mov	dh,0
	shl	dx,1
	add	ax,dx	

setcs2:	mov	cursor,ax
	pop	dx
	pop	ax
	cmp	al,al		;RZ
	return


comment |DUPCHR - ^D: duplicate char
|

dupchr:	mov	al,ctlstr+1	;al = char to duplicate
	mov	cl,ctlstr+2

dupch2:	mov	ah,attrib
	mov	ch,0		;cx = number of duplications
	add	csrcol,cl	;Update cursor column
	mov	di,cursor
	rep	stosw
	mov	cursor,di	;Update cursor
	cmp	al,al
	return


comment |EBLCRS - ^E: enable/disable hardware cursor
|

eblcrs:	mov	al,ctlstr+1	;al = code
	cmp	al,7fh		;Display hardware cursor?
	jz	short eblcr4	;Yes, go do it
	test	al,al
	js	short eblcr2	;Reset some flags
	or	dspflgs,al	;Set some flags
	cmp	al,al		;RZ, i.e., no char to display
	return

eblcr2:	and	dspflgs,al	;Reset some flags
	cmp	al,al
	return

eblcr4:	call	discrs
	cmp	al,al
	return


comment |BELL - ^G: gong! i.e., ring bell
|

bell:	mov	ah,0eh
	int	10h
	cmp	al,al
	return


comment |BCKSPC - ^H: destructive backspace
|

bckspc:	test	csrcol,0ffh	;Already at left margin?
	rz			;Yes, RZ
	dec	csrcol
	sub	cursor,2
	les	di,dcursor
	mov	al,' '
	mov	ah,attrib	;Attribute byte
	stosw
	cmp	al,al
	return


comment |TAB - ^I: tab
|

tab:	mov	al,8
	sub	al,csrcol
	and	al,7		;al = fraction of tab position
	xchg	cx,ax
	jnz	short blanks
	mov	cl,8		;Need full 8 cols. Fall thru to blanks

blanks:	mov	al,' '		;al = char to duplicate
	jmp	dupch2		;cl = # times to duplicate char


comment |LINEFD - ^J: linefeed
|

linefd:	mov	ax,lpp
	dec	ax
	cmp	al,csrrow	;Above screen bottom?
	ja	short linef6	;Yes, go increment cursor row and screen ptr
	inc	ax		;Cursor row and screen ptr remain unchanged
	mov	ctlhght,al	;Store hght = lpp
	mov	ax,linel
	shr	ax,1
	mov	ctlwdth,al	;Store wdth = linel/2
	mov	word ptr ctlrow,0	;Store row = column = 0
	call	scrlup		;Scroll whole screen up
	mov	di,cursor	;Erase row without changing csrcol or cursor
	mov	al,csrcol
	mov	ah,0
	shl	ax,1		;Convert to word offset
	sub	di,ax		;Start erasing at left column
	mov	cx,linel
	shr	cx,1		;All columns
	mov	ah,attrib
	mov	al,' '
	rep	stosw
	jmp	short linef8
linef6:	inc	csrrow
	mov	ax,linel
	add	cursor,ax
linef8:	cmp	al,al		;RZ
	return


comment |CLRSCN - ^L: clear lines 0 thru lpp-1. Leave cursor at (0,0)
|

clrscn:	xor	di,di
	mov	word ptr csrcol,di	;cursor row/col = 0
	mov	cursor,di
	mov	ax,lpp
	mov	cx,linel
	shr	cx,1
	mul	cl		;Leave dx alone
	xchg	cx,ax		;_cx = LINEL*lpp/2
	mov	al,' '
	mov	ah,attrib
	rep	stosw
	return


comment |CARETN - ^M: carriage return
|

caretn:	xor	ax,ax
	xchg	al,csrcol
	shl	ax,1
	sub	cursor,ax
	cmp	al,al
	return


comment |TOGUND - ^N: toggle underscore attribute
|

togund:	xor	attrib,UNDER
	cmp	al,al		;RZ
	return


comment |TOGRVV - ^R: toggle reverse-video attribute
|

togrvv:	xor	attrib,RVV
	cmp	al,al		;RZ
	return


comment |SCLRCT - ^S: scroll rectangle (col row wdth hght)
|

sclrct:	mov	dl,ctlcol	;dl = column
	mov	dh,ctlrow	;dh = row
	call	setcs1		;Store them in csrcol/csrrow and set cursor
	mov	cl,ctlwdth
	mov	ch,0		;cx = wdth
	test	ctlhght,0ffh	;Move rectangle up?
	jns	short sclrc2
	neg	ctlhght		;Make hght positive
	call	scrlup
	jmp	short sclrc8
sclrc2:	call	scrldn	
sclrc8:	cmp	al,al
	return


comment |SETCOL - ^X: set column (n)
|

setcol:	mov	al,ctlstr+1	;al = n
	xchg	al,csrcol
	sub	al,csrcol
	cbw			;ax = old column - new column
	shl	ax,1
	sub	cursor,ax
	cmp	al,al
	return


comment |SETROW - ^Y: set row (n)
|

setrow:	mov	al,ctlstr+1	;al = n
	xchg	al,csrrow
	sub	al,csrrow
	cbw			;ax = old row - new row
	imul	linel
	sub	cursor,ax
	cmp	al,al
	return


comment |WIPRCT - ^W: wipe (erase) rectangle (col row wdth hght)
|

wiprct:	mov	dl,ctlcol	;dl = column
	mov	dh,ctlrow	;dh = row
	call	setcs1		;Store them in csrcol/csrrow and set cursor
	mov	cl,ctlwdth
	mov	ch,0		;cx = wdth
	les	di,dcursor	;Point at upper-left corner of rectangle
	mov	al,' '
	mov	ah,attrib
wiprc2:	push	cx
	rep	stosw
	pop	cx
	add	di,linel	;Point at next row
	dec	ctlhght		;More rows to go?
	jnz	wiprc2		;Yes
	cmp	al,al
	return



;**************************Screen display routines****************************

comment |TOM - display message [si] at current screen postion
|

tom:	lodsb			;Display string [si] at es:[di]
	or	al,al		;Maybe.  Done?
	jnz	short tom2	;No
	cmp	al,byte ptr ctlcnt	;Yes, unless getting control seq args
	rz				;We're done
tom2:	call	co
	jmp	tom


comment	|VSTORE - store char al in video RAM at offset cursor with attribute
attrib.  Increment cursor, csrcol accordingly.  ax, di, es changed
|

blank:	mov	al,' '

vstore:	les	di,dcursor
	mov	ah,attrib	;Attribute byte
	stosw
	mov	ax,linel
	shr	ax,1		;ax = # cols on screen
	dec	ax
	cmp	al,csrcol	;At rightmost column?
	rbe			;Yes
	mov	cursor,di	;Update cursor and csrcol
	inc	csrcol
	return


comment |SCRLUP - Scroll current window up one line.  Window is defined by its
upper left column and row (ctlcol, ctlrow) and its width and height (ctlwdth,
ctlhght)

ax, cx changed.
|

scrlup:	mov	al,ctlcol	;al = left column of window
	cbw
	shl	ax,1
	xchg	si,ax		;si = offset on screen row
	mov	al,ctlrow	;al = window row
	mul	byte ptr linel	;ax = screen offset of top window row
	add	ax,si		;ax is destination ptr
	mov	si,ax
	add	si,linel	;si is source ptr

scrlu2:	xor	dx,dx
	mov	dl,ctlhght	;dx = # rows in window
	dec	dx		;dx = # rows to move
	jle	short scrlu8	;No rows to move

scrlu3:	xchg	di,ax		;di = destination ptr
	mov	ch,0
	mov	cl,ctlwdth	;cx = window width
	mov	ah,0
	mov	al,ctlcol	;ax = left column of window
	shl	ax,1
	add	si,ax
	add	di,ax
	mov	bx,linel
	sub	bx,cx
	sub	bx,cx		;bx = linel - 2*wdth = # extra columns in row
	cmp	si,di	;Up?
	ja	short scrlu4
	neg	bx		;No
scrlu4:	mov	es,vidram
	push	es
	pop	ds		;ds = es = vidram

scrlu6:	push	cx
	rep	movsw		;Move a row (or part thereof)
	cmp	cs:intmsg1,' '
;	callnz	serst1		;Buffer serial input unless interrupt driven
	pop	cx
	add	di,bx		;Add # columns in rest of line
	add	si,bx
	dec	dx		;More rows to go?
	jnz	scrlu6		;Yes
	push	cs
	pop	ds

scrlu8:	cld			;Ensure validity of SST's assumption that DF=0
	return


comment |SCRLDN - Scroll current window down one line. ax, cx changed.
|

scrldn:	xor	ax,ax
	mov	al,ctlhght	;al = window height
	add	al,ctlrow	;Add in starting row
	dec	ax		;ax = bottom row of window
	mul	linel		;ax = screen offset of bottom row; dx = 0
 	mov	dl,ctlwdth	;dl = window width
	add	dl,ctlcol	;Add in starting column
	dec	dx		;dx = ending column of window
	shl	dx,1		;Convert to word displacement for screen
	add	ax,dx		;ax = linel*(row+hght-1) + 2*(col+wdth-1)
	mov	si,ax
	sub	si,linel	;si -> line above
	std			;movsw downwards (cld before return)
	jmp	scrlu2		;Proceed as for scrlup


comment |VDINIT - initialize video parameters
|

vdinit:	push	ds		;This program writes the screen directly
	mov	ax,40h		; since speed is of the essence!
	mov	ds,ax
	assume	ds:bdata
	mov	ax,word ptr crtmode
	cmp	al,7		;Monochrome?
	jz	short vdini2	;Yes, all set up
	mov	vidram,0b800h	;No, color/graphics
	mov	crtprt,3d4h
	mov	al,crtrows	;Get crtrows (valid for EGA/VGA)
	test	al,al
	jz	short vdini2	;Not defined (must be ancient CGA!)
	inc	ax
	mov	byte ptr lpp,al
vdini2:	mov	al,ah		;Store # bytes per screen row
	mov	ah,0
	shl	ax,1
	mov	linel,ax
	mov	dx,cursorpos
	pop	ds
	assume	ds:cseg
	jmp	setcs1
