title	keytran.asm - prototype program to reconfigure keyboard

;(c) copyright 1994 by Scroll Systems, Inc.

CR	=	0dh
LF	=	0ah
EOI	=	20h	;End Of Interrupt command
kb_data	equ	60h	;Keyboard data port
kb_ctl	equ	61h	;Keyboard control port

data	segment	at 40h	;See PC Technical Reference BIOS listing for defs
	org	17h
kb_flag	label	byte	;RAM keyboard flag byte with meanings

;bit 0 - right shift key pressed
;bit 1 - left shift key pressed
;bit 2 - Ctrl key pressed
;bit 3 - Alt key pressed
;bit 4 - Scroll-Lock state on
;bit 5 - Num-Lock state on
;bit 6 - Caps-Lock state on
;bit 7 - Insert state on

	org	1ah
buffer_head	dw	?
buffer_tail	dw	?
kb_buffer	dw	16 dup(?)	;Only 16 entries, sigh!
kb_buffer_end	label	word
data	ends

intpag	segment	at 0
	org	4*9
int9	dw	?
intpag	ends


cseg	segment
	assume	cs:cseg

dvorak	db	?	;Dvorak flag
newcod	db	?	;Unmodified key code for other programs

	org	58h
oint9	dd	?	;Old INT 9 address
prog	dw	?
off	= 5ch - 103h	;prog - offset table

	org	100h

start:	jmp	setup	;Bypass tables and keyboard interrupt code

table	db	0,0,0	;Unshift translation table, put what you want
ctable	db	0,0,0	;Ctrl table, put what you want

;Dvorak keyboard translation table

tabled	db	16,39,17,44,18,46,19,80,20,89,21,70,22,71,23,67,24,82
	db	25,76,26,47,30,65,31,79,32,69,33,85,34,73,35,68,36,72
	db	37,84,38,78,39,83,40,45,44,59,45,81,46,74,47,75,48,88
	db	49,66,50,77,51,87,52,86,53,90,0

;Upper and lower case alternate translation table for umlaut chars

altlc	db	30,132,24,148,22,129,48,225,0
altuc	db	30,142,24,153,22,154,0


;Keyboard interrupt routine

keybrd:	push	ax	;Save registers used
	push	bx
	push	cx
	push	dx
	push	si
	push	di
	push	ds
	push	es
	cld
	sti
	mov	ax,DATA	;ROM BIOS RAM segment
	mov	ds,ax
	assume	ds:DATA
	in	al,kb_data	;Get and save key code
	mov	newcod,al
	cmp	dvorak,83	;Transparent mode?
	jnz	keybr0
	test	kb_flag,4	;Yes.  Check for Ctrl-+ only
	jz	keybr2
	cmp	al,78	;"+" key on numeric keypad?
	jnz	keybr2
	jmp	short keybr7	;Yes. Reenable translations

keybr0:	test	al,80h	;Break code?
	jz	keybr3

keybr2:	pop	es	;Use INT-9 routine active when symbol.com run
	pop	ds
	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	jmp	[oint9]

keybr3:	test	kb_flag,4	;Ctrl key pressed?
	jnz	keybr6
	test	kb_flag,8	;No. Alt key pressed?
	jnz	altran

	mov	di,offset table + off	;No. Translate key code
	cmp	dvorak,32	;Use Dvorak table?
	jnz	keybr4
	mov	di,offset tabled + off	;Yep
keybr4:	call	trans
	jnz	keybr2
	call	shfchk	;Translation occurred; shift to lower case?
	jnz	keybr5
	cmp	al,"A"	;Yes if alphabetic
	jc	keybr5
	cmp	al,"Z"
	ja	keybr5
	or	al,20h	;Convert to lower case
keybr5:	jmp	k61	;Store code

keybr6:	test	kb_flag,8	;Ctrl key. Alt key too?
	jz	keybr8
	cmp	al,16	;Yes. Querty control code?
	jz	keybr7
keyb62:	cmp	al,32	;No. Dvorak?
	jnz	keybr2
keybr7:	mov	dvorak,al	;Update keyboard control code
	jmp	k26	;Restore regs and iret

keybr8:	cmp	al,83	;Ctrl-Del?
	jz	keybr7
	mov	di,offset ctable + off
	jmp	short altra4

altran:	cmp	dvorak,83
	jz	altra2
	test	kb_flag,8	;Alt key pressed?
altra2:	jnz	altra3
	jmp	keybr2
altra3:	mov	di,offset altlc + off
	call	shfchk
	jz	altra4
	mov	di,offset altuc + off
altra4:	call	trans
	jz	k61
	jmp	keybr2	;No translation, use previously active INT-9 routine

k61:	mov	bx,buffer_tail	;Like IBM PC ROM code: store char in buffer
	mov	si,bx
	add	bx,2
	cmp	bx,offset kb_buffer_end
	jne	k612
	mov	bx,offset kb_buffer
k612:	cmp	bx,buffer_head
	je	k26	;Should give error beep: too much code to dup
	mov	[si],ax
	mov	buffer_tail,bx

k26:	cli		;Restore keyboard and return
	in	al,kb_ctl
	mov	ah,al
	or	al,80h
	jmp	$+2	;Slow down
	out	kb_ctl,al
	xchg	ah,al
	jmp	$+2
	out	kb_ctl,al
	mov	al,EOI	;Send End Of Interrupt command to 8259 interrupt
	out	20h,al	; controller
	pop	es
	pop	ds
	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	iret


comment |SHFCHK - RZ iff Caps-Lock not active and neither Shift key is
pressed.  No regs changed
|

shfchk:	test	kb_flag,40h	;Caps-Lock state on?
	jz	shfch2
	test	kb_flag,3	;Yes, either shift key pressed?
	jnz	shfch1
	or	sp,sp	;No: RNZ
	ret
shfch1:	cmp	ax,ax	;Yes: RZ
	ret

shfch2:	test	kb_flag,3	;Caps-Lock not on; either Shift key pressed?
	ret


comment |TRANS - Translate routine.  Replaces byte in string cs:[di] by byte
following matched byte.  RZ iff match found.
|

trans:	cmp	al,cs:[di]
	jz	trans2
	inc	di
	inc	di
	cmp	byte ptr cs:[di],0
	jnz	trans
	or	al,al
	ret		;RNZ
trans2:	mov	al,cs:[di+1]
	ret		;RZ

keyend:			;End of resident code


hlpinfo:

db "keytran 94. (c) copyright 1994 Scroll Systems, Inc.",CR,LF
db "This TSR program takes over the PC's keyboard interrupt (INT 9), ",CR,LF
db "translating some input key codes to new values, and passing others ",CR,LF
db "onto the INT-9 handler active when KEYTRAN was run.  Ctrl-Alt-D ",CR,LF
db "turns on a Dvorak keyboard layout; Ctrl-Alt-Q returns to qwerty. ",CR,LF
db "Alt aoubAOU give ",132,148,129,225,142,153,154
db ".  Ctrl-Del turns off all translations,",CR,LF
db "and Ctrl-+ (on numeric keypad) turns them back on.  This feature is",CR,LF
db "needed to allow KEYTRAN to be resident along with other keyboard ",CR,LF
db "handlers like communication programs that might assign different",CR,LF
db "meanings to the keys. See Section 9-1 for background on how the",CR,LF
db "program works.",CR,LF,'$'

setup:	mov	dx,offset hlpinfo	;Display sign on
	mov	ah,9
	int	21h
	xor	ax,ax	;Get old int 9 vector
	mov	es,ax
	assume	es:intpag
	mov	ax,int9
	mov	word ptr oint9,ax
	mov	ax,int9+2
	mov	word ptr oint9+2,ax
	mov	di,offset prog	;Move keyboard routine down to prog
	mov	si,offset table
	mov	cx,offset keyend - offset table
	push	ds
	pop	es
	rep	movsb	;(Saves space)
	push	di
	mov	dx,offset keybrd+off	;Set DS:DX to target INT 9 addr
	mov	al,9	;Set the keyboard interrupt (INT 9)
	mov	ah,25h	; to DS:DX via DOS function call
	int	21h
	pop	dx	;End program but leave keybrd resident
	int	27h

cseg	ends

	end	start
