title	PMTEST - program to check out protected mode

comment |Routine illustrating how to set up a GDT and an IDT, switch into
protected mode, display a message on the screen, and return to real mode.  You
can use SST to single step up to the call protmd, but then either type g at
the command line or break after returning from the call realmd.

|

public	cnv5, cputyp, cursor, defvid, dumint, eblA20, getcpu
public	pminit, pmtest, protmd, realmd
public	strdsb, strdsc, tom, vidseg, vstore

include	return.inc

NINT	=	32		;First 20h ints
IBFUL	=	2		;8042 Input buffer full
ENBL20	=	0dfh		;Enable Address line 20 code
DSBL20	=	0ddh		;Disable Address line 20 code
PORT_A	=	60h		;8042 port A
STSPRT	=	64h		;8042 status port
CMSPRT	=	70h		;CMOS port
SHUTDWN =	8fh		;CMOS Shut down register for 286


dscptr	struc			;Descriptor
sglen	dw	?		;Segment max length
sgbase	dw	?		;Segment base low word
sghbac	dw	?		;Segment access byte/base high byte
hglmbs	dw	?		;Reserved word
dscptr	ends

sst_gdt	struc			;SST compatible GDT
nuldsc	dq	?		;00 - Null selector descriptor not accessible
gdtdsc	dq	?		;08 - Global Descriptor Table descriptor
srcdsc	dq	?		;10 - Source Descriptor
tgtdsc	dq	?		;18 - Target Descriptor
symdsc	dq	?		;20 - SST symbol table descriptor
stkdsc	dq	?		;28 - SST Stack (and prefix) Descriptor
dgpdsc	dq	?		;30 - SST DGROUP Descriptor
sstdsc	dq	?		;38 - SST Code Descriptor = prefix + 10
bdtdsc	dq	?		;40 - BIOS Data descriptor
intdsc	dq	?		;48 - Real mode interrupt vector descriptor
idtdsc	dq	?		;50 - IDT descriptor
crtdsc	dq	?		;58 - CRT descriptor
ussdsc	dq	?		;60 - User ss descriptor
udsdsc	dq	?		;68 - User ds descriptor
ucsdsc	dq	?		;70 - User cs descriptor
usrdsc	dq	?		;78 - extra user descriptor
sst_gdt	ends

bdata	segment at 40h		;BIOS data segment
	org	49h		; p. A-2)
crtmode	db	?		;Current CRT mode
crtcols	db	?		;# cols on text screen
	org	50h
curpos	dw	?		;row/column
	org	67h		;Addresses needed for 286 shutdown
rominit	dw	?		;SHUT9 loads ss:sp from here
romseg	dw	?		; and ret's
bdata	ends

dseg	segment
gdt	sst_gdt	<>		;GDT compatible with SST
idt	dq	NINT dup(?)	;Protected-mode IDT
rmitbl	df	0ffffh		;Real-mode interrupt vector table limit
gdtadr	df	?		;GDT limit and base address for lgdt
idtadr	df	?		;IDT limit and base address for lgdt
cursor	dw	0		;Cursor offset in video RAM
vidseg	dw	?		;Video RAM segment/selector
dcursor	equ	dword ptr cursor
cputyp	db	?		;0-5 for 80x86
msg	db	"pmtest 94. (c) copyright 1994 Scroll Systems, Inc.",0dh,0ah
	db	"Program to check out protected mode $"
cpumsg	db	0dh,0ah,"Need an 80286 or later CPU$"
pmmsg	db	0dh,0ah,"Already running in protected mode "
	db	"(must be Virtual-86 mode)",0dh,0ah
	db	"Be sure that emm386.sys isn't active and that "
	db	"you're not in a Windows DOS box$"
runmsg	db	" Running in protected mode ",1,' ',0
OKmsg	db	0dh,0ah,"Switched back to real mode$"
dseg	ends

sseg	segment	stack
	dw	100 dup(?)
sseg	ends

cseg	segment
	assume	cs:cseg
.386p


pmtest:	mov	ax,dseg		;Point at data area
	mov	ds,ax
	mov	es,ax
	assume	ds:dseg

	mov	dx,offset msg	;Display program message
	mov	ah,9
	int	21h
	call	defvid		;Define real-mode video segment
	mov	dx,offset pmmsg
	smsw	ax
	test	al,1		;Already running in protected mode?
	jnz	short pmtes8	;Yes, can't run (or maybe try for DPMI)
	call	getcpu		;Get al = CPU type
	cmp	al,2		;80286 or bigger?
	mov	dx,offset cpumsg
	jc	short pmtes8	;No, can't run
	mov	cputyp,al	;Save type for other routines to check
	call	pminit		;Initialize GDT and IDT
	call	protmd		;Switch into protected mode
	mov	si,offset runmsg ;Display run msg at screen offset
	call	tom		; given by cursor
	call	realmd		;Switch back to real mode
	mov	dx,offset OKmsg

pmtes8:	mov	ah,9		;Display success or failure message
	int	21h
	mov	ah,4ch		;Return to MS-DOS
	int	21h


;Define video-RAM real mode segment. ax, bx changed

defvid:	push	es
	mov	ax,bdata
	mov	es,ax
	assume	es:bdata
	cmp	crtmode,7	;Monochrome?
	mov	ax,0b000h
	jz	short defvi2	;Yes
	mov	ah,0b8h		;Color/graphics
defvi2:	mov	vidseg,ax
	mov	bx,curpos	;bh/bl = DOS row/column
	mov	al,crtcols	;al = # columns on screen
	mul	bh
	mov	bh,0
	add	ax,bx		;Add in columns
	shl	ax,1		;Convert byte offset to word offset
	mov	cursor,ax
	pop	es
	return


;Switch from real mode to protected mode

protmd:	cmp	cputyp,2	;Check on x in x86 (see getcpu)
	rc			;Has to be at least a 286
	smsw	ax		;Get current machine status word
	test	al,1		;Already in protected mode?
	rnz			;Yes, return (might be in a DOS box)
	call	eblA20			;Enable Address line 20
	lgdt 	fword ptr gdtadr	;Load gdtr
	cli				;No interrupts until ss setup in pm
	lidt 	fword ptr idtadr	;Load idtr

;Unmask hardware interrupts you have handlers for, e.g., for keyboard
;and clock, output 0fch to port 21h and 0ffh to port 0a1h.

	smsw	ax		;(low word of cr0)
	or	al,1		;Set PM bit
	lmsw	ax		;Switch into protected mode
	db	0eah		;Far jump (never found a way to get asm
	dw	protm4		; to do this).  jmp clears prefetch queue
	dw	ucsdsc		; and sets up cs in protected mode
protm4:	mov	ax,ussdsc	;Setup stack segment 
	mov	ss,ax		; with same sp as real mode
	sti
	mov	ax,udsdsc	;Setup data segment
	mov	ds,ax
	mov	es,ax
	mov	vidseg,crtdsc	;Setup selector for direct-to-screen functions
				;Define any other selectors you need
	return			;Return in protected mode


;Switch back to real address mode. Note that most of the code is for the
;286 case

realmd:	cli
	mov	ax,sseg		;ax = real-mode stack segment
	cmp	cputyp,3	;386 or bigger?
	jae	short realm1	;Yes, ...and the livin' is easy
	lidt	fword ptr rmitbl+2 ;Load up IDTR with 0-length IDT
	push	bdata		;80286 shutdown.
	pop	ds		;Jump through hoops!
	assume	ds:bdata	;ds -> BIOS RAM
	mov	romseg,ax	;Save real-mode stack segment
	mov	al,SHUTDWN	;Output shutdown return code
	out	CMSPRT,al	;Also turns off NMI
	pushf			;Push iret frame that returns to realm5 
	push	cseg		; below
	push	offset realm5
	pusha			;Save regs for BIOS shut9 
	push	es		;Finish pushing registers for shut9
	push	ds
	mov	rominit,sp	;Save current sp for shut9
	mov	al,9		;Shutdown option to return via ss:sp
	out	CMSPRT+1,al	; given by dword ptr rominit
	int	3		;Interrupt thru 0-length interrupt table 
				;Causes "triple fault", resetting x86

	assume	ds:dseg		;80386 easy way to do things
realm1:	xchg	si,ax		;Save real-mode stack segment
	lidt	fword ptr rmitbl ;Load real-mode IDT
	mov	eax,cr0		;For shutdown, need whole cr0
	and	eax,7ffffffeh	;Kill paging and protected-mode bits
	mov	cr0,eax		;Back to real mode
	db	0eah		;Far jump. Some assemblers can take
	dw	realm3		; jmp far ptr realm3 if realm3 is a far label
	dw	cseg
realm3:	mov	ss,si		;Define real-mode stack. sp stays same


;80286 returns here after reset with ds = bdata
;ss:sp = real-mode version of value on entry to realmd.

realm5:	mov	ax,dseg		;Restore various segments to real mode 
	mov	ds,ax		; values
	mov	es,ax
	call	defvid		;Set vidseg = real-mode video RAM segment
	sti
	return			;Return in real mode to caller of realmd


;pminit - initialize GDT to point to current data and code segments


pminit:	pushf			;Kill possible Nested Task flag
	pop	ax		; or first iret will bomb
	and	ax,3fffh
	push	ax
	popf
	mov	ax,ds		;Setup descriptors
	mov	bx,offset gdt	;gdtdsc
	mov	si,gdtdsc
	call	strdsc		;Base address = ds:gdtdsc
	mov	[si].sglen,size sst_gdt-1	;Adjust GDT descriptor length
	mov	di,offset gdtadr	;Copy gdt limit and address for lgdt
	movsw			;Limit
	movsw			;Low word of base address
	movsb			;Base address byte 2

	mov	ax,bdata	;Bimodal BIOS data segment
	mov	si,bdtdsc	; in case you want to define clock/kbd ints
	call	strdsb		;Base address = 400h
	mov	ax,vidseg	;Setup video-RAM descriptor
	mov	si,crtdsc
	call	strdsb
	mov	ax,ss		;Stack-segment descriptor
	mov	si,ussdsc
	call	strdsb
	mov	ax,ds		;Data-segment descriptor
	mov	si,udsdsc
	call	strdsb
	mov	ax,cs		;Code-segment descriptor
	mov	si,ucsdsc
	call	strdsb
	mov	byte ptr [si+1].sghbac,9ah	;Change to code descriptor

	mov	ax,ds		;Define interrupt descriptor
	mov	bx,offset idt
	mov	si,idtdsc
	call	strdsc		;Base address = ds:sstidt
	mov	[si].sglen,8*NINT-1	;IDT length includes NINT entries

	mov	di,offset idtadr	;Copy idt limit and absolute address
	movsw				; over to idtad for lidt instruction
	movsw
	movsb

	mov	di,offset idt	;Define protected-mode IDT
	mov	cx,NINT		;Initialize first NINT INT's to dummy 
pmini2:	mov	ax,offset dumint	; interrupt handler
	stosw			;(Point to actual interrupt handlers if
	mov	ax,ucsdsc	; you define them)
	stosw
	mov	ax,8600h	;80286 ring-0 interrupt gate. Could use
	stosw			; 80386 interrupt or task gates (handlers
	xor	ax,ax		; then need iretd instead of iret)
	stosw
	loop	pmini2
	return

dumint:	iret			;Add appropriate handlers according to need


;Enable (ah = ENBL20) or disable (ah = DSBL20) A20 line

eblA20:	cli			;Enter with ah = ENBL20 or DSBL20
	call	eblA22		;Wait till 8042 not busy
	rnz			;No good... RNZ
	mov	al,0d1h		;8042 command to write output port
	out	STSPRT,al	;Output to 8042
	call	eblA22		;Wait till it does its thing
	rnz			;Whoops
	mov	al,ah		;Output ENBL20 or DSBL20
	out	PORT_A,al
eblA22:	xor	cx,cx		;Wait for 8042 not busy
eblA24:	in	al,STSPRT
	and	al,IBFUL	;Still busy?
	loopnz	eblA24		;Yes; times out eventually
	return


comment |STRDSB - Create GDT data descriptor for real-mode address ax:0 and
selector si.  Return *si = descriptor.  ax, bx, si changed
|

strdsb:	xor	bx,bx		;Offset = 0


comment |STRDSC - Create GDT data descriptor for real mode address ax:bx and
selector si.  Return *si = descriptor.  ax, bx, si changed
|

strdsc:	call	cnv5		;Convert ax:bx to flat 5-digit value


comment |STRDSX - Create GDT for flat address axbx and selector si.  Return
*si = descriptor.  ax, bx, si changed
|

strdsx:	add	si,offset gdt
strds1:	mov	ah,92h		;Set ring-0, r/w access, not yet accessed


comment |STRDS2 - store descriptor *si with length 0ffffh, base albx, and
attribute byte ah.  No regs changed.
|

strds2:	mov	[si].sglen,-1
	mov	[si].sgbase,bx
	mov	[si].sghbac,ax
	mov	[si].hglmbs,0
	return


comment |CNV5 - Convert real-segment ax:bx into 5 hex-digit addr and RC.  ax,
bx changed.
|

cnv5:	rol	ax,4		;Shift 3 low nibbles left for offset
	push	ax		;Save high nibble now in low position
	and	al,0f0h		;Kill it for offset
	add	bx,ax		;Add in offset
	pop	ax		;Recover high nibble
	adc	ax,0		;Add in possible CF
	and	ax,0fh		;Kill low nibbles (now in high positions)
	return


;tom - display message [si] at current screen postion

tom:	lodsb			;Display string [si] at es:[di]
	or	al,al		;Maybe.  Done?
	rz			;We're done
tom2:	call	vstore
	jmp	tom


;vstore - store char al in video RAM at offset given by di

vstore:	les	di,dcursor
	mov	ah,70h		;Attribute byte
	stosw
	mov	cursor,di	;Update cursor
	return


;getcpu works with 16-bit or 32-bit flat mode.  Define _ax, etc. for 16-bit

_ax	equ	ax
_bx	equ	bx
_cx	equ	cx
_dx	equ	dx
_si	equ	si
_di	equ	di
_bp	equ	bp
_sp	equ	sp
_es	equ	es
_ss	equ	ss

include	getcpu.asm

cseg	ends

	end	pmtest
