title	RBMODE - program to check out real-big mode

comment |Routine illustrating how to set up a tiny GDT switch into protected
mode just long enuf to change the data-segment limit to 4GB, return to real
mode, and use large data segment offsets.  You can use SST to single step
everything except the small portion of code leading up to fxsgl2
|

public	chkXMS, eblA20, fxsgl1, fxsglm, gdt, gdtadr, getcpu
public	gtxli1, gtxlim, hRAM, intd
public	XMSctrl, XMSvern, xramin, xramax


include	return.inc

IBFUL	=	2		;8042 Input buffer full
ENBL20	=	0dfh		;Enable Address line 20 code
DSBL20	=	0ddh		;Disable Address line 20 code
PORT_A	=	60h		;8042 port A
STSPRT	=	64h		;8042 status port

vdskseg	segment at 0		;Device driver segment given by 
	org	12h		; INT19+2
vdisk	dw	?		;"VDISK"
	org	2ch
xramtop	dw	?		;Top of Xtended RAM
vdskseg	ends

intseg	segment	at 0
	org	4*19h
int19	dw	?
intseg	ends

;Global Descriptor Table for setting up full memory access in real mode

dseg	segment public
gdt	dw	0ffffh		;00 - Null selector descriptor
	dw	3 dup(0)	;Use for gdt absolute address for lgdt
gigadsc	dw	0ffffh		;08 - 4-gigabyte data seg descriptor
	dw	0		;Absolute address unimportant for limit
	db	0		; change
	db	92h		;Attribute for a present data seg
	dw	8fh		;Access to full 4 gigabytes

gdtadr	df	?		;Place for sgdt to store gdt
xramin	dd	?		;Absolute address of starting xram
xramax	dd	?		;Absolute address of ending xram
XMSvern	dw	?		;XMS version number
XMSctrl	dd	?		;Address of XMS handler
XMSfree	dw	0		;#K in largest XMS XRAM block
oneK	dw	1024
hRAM	dw	?		;XMS XRAM block handle
msg	db	"rbmtest 94. (c) copyright 1994 Scroll Systems, Inc.",0dh,0ah
	db	"Program to check out real-big mode",0dh,0ah,'$'
cpumsg	db	"Need an 80386 or later CPU",0dh,0ah,'$'
pmmsg	db	"Already running in protected mode "
	db	"(must be Virtual-86 mode)",0dh,0ah
	db	"Be sure that emm386.sys isn't active and that "
	db	"you're not in a Windows DOS box$"
rbmsg	db	"The 5 chars starting at absolute RAM location 100003h = $"
OKmsg	db	0dh,0ah,"Restored allocated XRAM and previous int-13 handler$"
dseg	ends


sseg	segment	stack
	dw	100 dup(?)
sseg	ends

cseg	segment
	assume	cs:cseg
.386p

intds	dd	?		;Saved int 13 (0dh) value (in cseg)

rbmode:	mov	ax,dseg
	mov	ds,ax
	mov	es,ax
	assume	ds:dseg

	mov	dx,offset msg	;Display program message
	mov	ah,9
	int	21h
	call	getcpu		;Get al = CPU type
	cmp	al,3		;80386 or bigger?
	mov	dx,offset cpumsg
	jc	short rbmod9
	call	fxsglm		;Try to fix the segment limits
	mov	dx,offset pmmsg
	jnz	short rbmod9	;Can't run in V86 mode
	call	gtxlim		;Define xramax and xramin absolute addresses
				;You can use RAM from xramin to xramax
	mov	dx,offset rbmsg
	mov	ah,9
	int	21h		;Display RAM at absolute location 100000h
	push	ds

	push	0		;As demo, print out 5 characters starting at
	pop	ds		; absolute location 100003h
	mov	esi,100003h
	mov	cx,5
rbmod5:	lods	byte ptr [esi]
	xchg	dx,ax
	mov	ah,2
	int	21h
	loop	rbmod5


;Restore previous int-13 handler and release allocated XRAM

	lds	dx,intds	;Restore previous int-13 handler
	mov	ax,250dh
	int	21h
	pop	ds
	mov	dx,hRAM		;Free XMS RAM block if allocated
	or	dx,dx
	jnz	short rbmod6
	mov	ah,0dh		;Yes, unlock it
	call	[XMSctrl]
	mov	ah,0ah		;Free it
	call	[XMSctrl]
rbmod6:	mov	dx,offset OKmsg

rbmod9:	mov	ah,9		;Display success or failure message
	int	21h
	mov	ah,4ch		;Return to MS-DOS
	int	21h


;fxsglm - fix ds and es segment limits to access the full RAM

fxsglm:	smsw	ax		;Determine operating mode
	test	al,1		;Running in protect mode?
	jnz	short fxsgl9	;Yes, can't use real big mode
	mov	ax,350dh	;Get es:bx = current int-0dh handler
	int	21h		; address
	mov	ah,25h		;Take over int 0dh
	mov	dx,offset intd
	push	ds
	push	cs
	pop	ds		;New interrupt vector = ds:dx
	int	21h
	pop	ds
	mov	word ptr intds,bx ;Save old int vector for restoration
	mov	word ptr intds+2,es ; on quitting
	xor	eax,eax		;Setup to store absolute GDT address
	mov	ax,ds
	shl	eax,4		;Shift left one nibble
	add	eax,offset gdt
	mov	dword ptr gdt+2,eax ;Save address for LGDT


;Load GDT register and switch into protected mode.  Called here on 
;int 0dh to re-fix segment limit if changed back by some other program

fxsgl1:	push	ds		;Save current segment-register values
	push	es
	mov	ax,dseg
	mov	ds,ax		;Load ds in case entered from intd
	assume	ds:dseg		;Could avoid by storing gdt in cseg
	lgdt	fword ptr gdt	;Load our mini GDT
	mov	ah,ENBL20	;Enable CPU address line 20
	call	eblA20		;(we never disable it)
	mov	cx,offset gigadsc - offset gdt
	mov	eax,cr0		;Get current cr0
	or	al,1		;Set PM bit
	cli			;No interrupts while in protected mode

;Don't single-step from here up to fxsgl2

	mov	cr0,eax		;Switch to protected mode
	jmp	$+2		;Flush prefetch queue (may not need)
	mov	ds,cx		;Turn on ds and es segment-length and
	mov	es,cx		; granularity bits. Could also fix fs and gs
	dec	ax		;Turn off PM bit
	mov	cr0,eax		;Return to real mode
	jmp	$+2		;Flush prefetch queue (may not need)

fxsgl2:	sti			;You can breathe now
	pop	es		;Restore real-mode register paragraphs
	pop	ds		; (granularity bits and limits remain set)
	cmp	ax,ax		;RZ to indicate success
fxsgl9:	return


;Real-mode int-0dh handler

intd:	push	bp		;Come here on fault due to attempted
	mov	bp,sp		; access beyond segment limit
	push	eax
	mov	ax,cs		;Our code interrupted?
	cmp	ax,[bp+4]
	jnz	short intd6	;No, go use previous int-0dh handler
	push	cx
	call	fxsgl1		;Restore large segment limit
	pop	cx
	pop	eax
	pop	bp
	iret
intd6:	pop	eax		;Use previous int-0dh handler
	pop	bp
	jmp	[intds]


;Define XRAM limits xramax and xramin.  ax, dx changed.

gtxlim:	call	chkXMS		;XMS there?
	jz	short gtxli1	;No, go try for older methods
	xchg	dx,ax
	mov	ah,9
	call	[XMSctrl]	;Ask for it
	shr	ax,1		;ax = 0 means it failed
	rnc			;Return failure
	mov	hRAM,dx		;Succeeded: store RAM handle
	mov	ah,0ch		;Lock it in place
	call	[XMSctrl]
	shr	ax,1		;RNC if failure
	rnc
	mov	word ptr xramin,bx ;Store absolute start address
	mov	word ptr xramin+2,dx
	movzx	eax,XMSfree	;Clear eax high word
	shl	eax,10h		;Convert from K to bytes
	add	eax,xramin
	mov	xramax,eax
	stc			;RC, i.e., success
	return

;Old int-15h and VDISK extended memory allocation methods

gtxli1:	mov	ah,88h		;Get ax = #K of extended RAM
	int	15h
	or	ax,ax		;None available
	rz
	add	ax,400h		;Add in extended RAM base (1024K)
	mul	oneK		;Convert #K into # bytes above 1 Meg
	sub	ax,1		;1 less
	sbb	dx,0		;Store in xramax (subsequent calls
	mov	word ptr xramax,ax	; get info from xramax)
	mov	word ptr xramax+2,dx
	push	es
	mov	dx,10h		;Default dxax = 100000H (1MB)
	xor	ax,ax
	mov	es,ax
	assume	es:INTSEG	;Real-mode interrupt table
	mov	es,int19+2
	assume	es:vdskseg
	cmp	vdisk,"DV"	;VDISK? 
	jnz	short gtxli2
	cmp	vdisk+2,"SI"
	jnz	short gtxli2
	mov	ax,xramtop	;Yes. Get current bottom of free XRAM
	mov	dl,byte ptr xramtop+2
gtxli2:	pop	es
	cmp	dx,word ptr xramax+2	;dxax >= xramax
	jnz	short gtxli4
	cmp	ax,word ptr xramax
gtxli4:	rnc
	mov	word ptr xramin,ax	;Save it for later allocations
	mov	word ptr xramin+2,dx
	return


;Enable (ah = ENBL20) or disable (ah = DSBL20) A20 line

eblA20:	cli			;Enter with ah = ENBL20 or DSBL20
	call	eblA22		;Wait till 8042 not busy
	rnz			;No good... RNZ
	mov	al,0d1h		;8042 command to write output port
	out	STSPRT,al	;Output to 8042
	call	eblA22		;Wait till it does its thing
	rnz			;Whoops
	mov	al,ah		;Output ENBL20 or DSBL20
	out	PORT_A,al
eblA22:	xor	cx,cx		;Wait for 8042 not busy
eblA24:	in	al,STSPRT
	and	al,IBFUL	;Still busy?
	loopnz	eblA24		;Yes; times out eventually
	return


;If XMS driver is installed, get its entry address, and biggest RAM block 
;available

chkXMS:	mov	ax,4300h	;Check to see if XMS driver is
	int	2fh		; installed
	cmp	al,80h		;Driver there?
	rnz			;No: RNZ
	push	es
	mov	ax,4310h	;Get XMS driver address
	int	2fh
	mov	word ptr XMSctrl,bx	;Store it for XMS calls
	mov	word ptr XMSctrl+2,es
	pop	es
	xor	ax,ax		;ah = 0 gets the XMS version #
	call	[XMSctrl]	;Call XMS driver
	mov	XMSvern,ax	;Save version number
	mov	ah,8		;Find out size of biggest XRAM
	call	[XMSctrl]	; block available
	mov	XMSfree,ax	;# KB of block
	cmp	ax,ax
	return			;RZ to signal XMS active


;getcpu works with 16-bit or 32-bit flat mode.  Define _ax, etc. for 16-bit

_ax	equ	ax
_bx	equ	bx
_cx	equ	cx
_dx	equ	dx
_si	equ	si
_di	equ	di
_bp	equ	bp
_sp	equ	sp
_es	equ	es
_ss	equ	ss

include	getcpu.asm

cseg	ends

	end	rbmode
