title	"ScanCode program - displays scancodes on screen"

;(c) copyright 1994 by Scroll Systems, Inc.

kb_data	equ	60h		;Keyboard data port
ESCAPE	=	27		;Quit code

bseg	segment	at 40h		;BIOS RAM segment
	org	49h
crtmode	db	?		;Current CRT display mode
crtcols	db	?		;# of screen columns
	org	84h
crtrows	db	?		;# of screen rows - 1 (EGA/VGA/XGA)
bseg	ends

cseg	segment
	assume	cs:cseg, ds:cseg

	org	80h		;Convenient place in PSP for some variables
pint9	dd	?		;Previous int-9 interrupt vector
vidoff	dw	?		;Screen offset
vidseg	dw	?		;Screen segment
vidptr	equ	dword ptr vidoff
linel	dw	?		;# bytes per screen row
lpp	dw	?		;# rows per screen
vidmax	dw	?		;# bytes per screen = lpp*linel
attrib	db	?		;Current screen attribute
lastcod	db	?		;Previous code

	org	100h		;.com file

start:	call	vdinit		;Initialize screen parameters
	call	blkscn		;Blank screen and display how-to-quit message
	mov	dh,byte ptr lpp	;Move BIOS cursor to end of msg
	dec	dh		;dh = last row on screen
	mov	dl,NMSG		;dl = column #
	mov	bl,0		;Screen page 0
	mov	ah,2		;Set cursor position
	int	10h

	mov	ax,3509h	;Get previous keyboard interrupt
	int	21h
	mov	word ptr pint9,bx	;Save it
	mov	word ptr pint9+2,es
	mov	dx,offset kbdint
	mov	ah,25h		;Set int-9 interrupt vector to
	int	21h		; point at our kbdint routine

start4:	mov	ax,0		;Wait for Esc key, eating all other keystokes
	int	16h
	cmp	al,ESCAPE	;Esc key?
	jnz	start4		;No, keep looking

	lds	dx,pint9	;Got Esc.  Load previous keyboard
	mov	ax,2509h	; interrupt vector and restore it
	int	21h
	ret			;Return to MS-DOS via int 20h in PSP
	

;Keyboard interrupt routine

	assume	ds:nothing	;Force cs override in following
.186
kbdint:	push	es		;Save all registers that we change
	pusha
	in	al,kb_data	;Get key code
	cmp	al,0e0h		;Alternate-key prefix?
	jz	short kbdin2	;Yes, just store it
	push	ax
	call	dspcod		;Display it
	pop	ax
kbdin2:	mov	lastcod,al	;Update last code value
	popa
	pop	es		;Chain to previous interrupt, which
	jmp	[pint9]		; resets keyboard & interrupt controllers


;Direct-to-screen-RAM display routines

dspbyt:	push	ax		;Display byte as pair of hex nibbles
	shr	al,4		;High nibble first
	call	dspnbl
	pop	ax		;Then low nibble

dspnbl:	and	al,0fh		;Convert nibble to ASCII
	add	al,90h		;Add 90H using BCD arithmetic
	daa
	adc	al,40h		;Then 40H plus CY
	daa			;Fall thru to vstore

vstore:	mov	ah,attrib	;ah = attribute byte
	stosw			;Store char code/attribute in screen RAM
vstor9:	ret


;Display make/break codes with possible 0e0h prefix in normal/reverse video

dspcod:	les	di,vidptr	;Point at screen RAM
	mov	ah,7		;Default normal video: gray on black
	test	al,al		;Sign bit set?
	jns	short dspco1	;No, it's a make code
	mov	ah,70h		;It's a break code: use reverse video
dspco1:	mov	attrib,ah	;Store attribute
	cmp	lastcod,0e0h	;Last code = 0e0h?
	jnz	short dspco4	;No
	push	ax		;Display 0e0h alternate-code prefix
	mov	al,0e0h		; with appropriate attribute
	call	dspbyt
	call	cndblk		;Display blank unless at start of row
	pop	ax
dspco4:	call	dspbyt		;Display al as a pair of hex nibbles
	mov	attrib,7	;Restore normal video
	call	cndblk		;Display blank unless at start of row
	mov	vidoff,di	;Update screen pointer
	cmp	di,vidmax	;Full screen?
	jc	vstor9		;No, go return

blkscn:	mov	cx,vidmax	;Blank screen: 80 columns wide
	shr	cx,1
	mov	vidoff,0	;Set origin to upper-left screen corner
	les	di,vidptr
	mov	ax,720h		;Blanks
	rep	stosw		;Zap!

	sub	di,linel	;Back up one row
	mov	si,offset msg	;Display how-to-quit message at screen bottom

dspmsg:	lodsb			;Display string [si] at es:[di]
	test	al,al		;End of string?
	jz	vstor9		;Yes, go return
	stosw			;Store char code/attribute in screen RAM
	jmp	dspmsg

cndblk:	mov	ax,di		;Display blank unless at start of row
	cwd
	div	linel		;dx = offset on row
	test	dx,dx		;If at beginning of row,
	jz	vstor9		; return; else fall thru to blank

blank:	mov	al," "		;Display ' '
	jmp	vstore


;Initialize screen parameters for direct screen storing

	assume	ds:cseg
vdinit:	mov	ax,bseg		;Point at BIOS RAM segment
	mov	es,ax
	assume	es:bseg
	mov	ax,0b000h	;Default monochrome display adapter (MDA)
	xor	bx,bx		;# of rows not decided yet
	cmp	crtmode,7	;MDA?
	jz	short vdini2	;Yes
	mov	ah,0b8h		;Make that CGA/EGA/VGA
	mov	bl,crtrows	;If EGA/VGA, crtrows is defined (else 0)
vdini2:	mov	vidseg,ax	;Store screen-RAM segment
	test	bl,bl		;# rows - 1 given?
	jnz	short vdini4	;Yes, use it
	mov	bl,24		;Assume 25 rows on screen
vdini4:	inc	bx
	mov	lpp,bx		;Store # rows on screen
	xor	ax,ax		;Set up vidptr to point at top of screen
	mov	vidoff,ax
	mov	al,crtcols	;al = # columns per screen row
	shl	ax,1
	mov	linel,ax	;linel = # bytes per screen row
	mul	bx		;ax = lpp * linel
	mov	vidmax,ax	;vidmax = # bytes per screen
	ret

msg	db	"scancode 94. (c) copyright 1994 Scroll Systems, Inc. "
	db	"Type Esc key to quit",0
NMSG	equ	$ - msg - 1

cseg	ends

	end	start
