title  LISTER - SIMPLE SERIAL PRINTER DRIVER
.186

;Serial port parameters

mctrl	equ	4		;Modem control port
lstat	equ	5		;Line Status port
mstat	equ	6		;Modem Status port
THRE	equ	20h		;Transmitter Holding Reg Empty
DSR	equ	20h		;Data Set Ready bit in modem status
UPARM	equ	0e3h		;9600 baud, 8 bits, no parity
				; (see BIOS listing for INT 14H in
				; the technical reference manual)
bdata	segment	at 40h		;BIOS RAM segment
bdata	ends

cseg	segment			;This coincides with the PSP
	assume	cs:cseg,ds:cseg	; since we generate a .com file

	org	5dh		;File Control Block 1 (see Sec. 4-7)
fcb1	db	?		;Easy place to get parameter from
				; command line
	org	0f0h
oint17h	dd	?		;Place in PSP for old int 17h vector
serprt	dw	?		;Serial port
portno	dw	?		;Serial port number

	org	100h		;For com files,
start:	jmp	setup		; execution begins at cs:100h

lister:	sti			;No need to keep interrupts disabled
	cmp	ah,1		;Check function code in ah
	jc	short print	;Print char in al if ah = 0
	jz	short init	;Initialize serial port if ah = 1
	jmp	cs:oint17h	;Use previous int vector for other functions

init:	push	ax		;ah = 1 initialization entry
	push	dx		;Save registers used
	mov	ax,UPARM	;Get UART parameters
	mov	dx,cs:portno	;Get serial port card #
	int	14h		;Initialize the UART
	mov	dx,cs:serprt
	add	dx,mctrl	;Set DTR and RTS low
	mov	al,3
	out	dx,al

done:	pop	dx		;Clean up and return
	pop	ax
	mov	ah,90h		;Force UART STATUS to be OK
	iret

print:	push	ax		;ah = 0 print entry
	push	dx
	push	ax		;Save output char
	mov	dx,cs:serprt
print1:	call	sersts		;ah/al = Modem/Line Status
	xor	ax,DSR*256+THRE	;Loop till DSR true & Transmitter
	test	ax,DSR*256+THRE	; Holding Register Empty
	jnz	print1
	pop	ax		;Restore char
	out	dx,al		;Ship it
	jmp	done

sersts:	push	dx		;Save serial data port number
	add	dx,mstat	;Point at modem status port
	in	al,dx
	mov	ah,al		;ah = modem status
	jmp	$+2		;Slow down for slow chips
	dec	dx		;Point at line status port
	in	al,dx		;al = line status
	pop	dx
	ret

;TSR setup code discarded after use

setup:	mov	al,fcb1		;Tell 'em which com port
	cmp	al,' '
	jz	short setup2	;None specified; leave default of com1
	cmp	al,'1'		;Must be 1 through 4
	jc	short setup3	;Bungo! Go announce the bad news
	cmp	al,'4'
	ja	short setup3	;Likewise
	mov	signon1,al	;Update choice
	dec	ax		;Convert com1/com2/com3/com4 to 0, 1, 2, 3
setup2:	and	ax,3		;Must be in range 0 - 3
	mov	portno,ax	;Save port number for init
	shl	ax,1		;Convert to word offset
	xchg	bx,ax		;bx = serial port table offset
	mov	ax,bdata	;Get desired serial port from BIOS RAM
	mov	es,ax
	mov	ax,es:[bx]
	mov	serprt,ax	;Save it
	or	ax,ax		;Port defined?
	mov	ah,9		;(int 21h display string function)
	jnz	short setup4	;Yes

setup3:	mov	dx,offset noscpt
	int	21h		;Display "No such port"
	mov	ax,4c01h	;Return to DOS with error 1
	int	21h		;Bye

setup4:	mov	dx,offset signon
	int	21h		;Say hi to the folks
	mov	ax,3517h	;Get current printer interrupt vector
	int	21h		; and save it for functions we don't handle
	mov	word ptr oint17h+2,es
	mov	word ptr oint17h,bx
	mov	dx,offset lister
	mov	ax,2517h	;Set the printer I/O interrupt (INT 17H)
	int	21h		; to DS:DX via DOS function call
	mov	ah,1		;Use the lister routine to initialize
	int	17h		; the serial port

	mov	dx,offset setup+0fh
	shr	dx,4		;dx = # paragraphs desired
	mov	ah,31h
	int	21h		;Terminate and Stay Resident

noscpt	db	"No such port$"
signon	db	"Print output to com"
signon1	db	"1$"

cseg	ends
	end	start
